package Pcore::Handle::Base;

use Pcore -role;

has uri => ( required => 1 );    # InstanceOf ['Pcore::Util::URI']

1;
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::Handle::Base

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 ATTRIBUTES

=head1 METHODS

=head1 SEE ALSO

=cut
