package Pcore::Util::Result::Class;

use Pcore -class;

with qw[Pcore::Util::Result::Role];

1;
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::Util::Result::Class

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 ATTRIBUTES

=head1 METHODS

=head1 SEE ALSO

=cut
