package Pcore::Util::URI::tor;

use Pcore -class;

extends qw[Pcore::Util::URI];

has default_port => 9050;
has is_tor       => 1;

1;
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::Util::URI::tor

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 ATTRIBUTES

=head1 METHODS

=head1 SEE ALSO

=cut
