package Pcore::Util::URI::sftp;

use Pcore -class;

extends qw[Pcore::Util::URI];

has is_ftp       => 1;
has is_secure    => 1;
has default_port => 22;

1;
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::Util::URI::sftp

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 ATTRIBUTES

=head1 METHODS

=head1 SEE ALSO

=cut
