package Pcore::Core::CLI::Cmd;

use Pcore -class;

sub CLI ($self) {
    return;
}

# return error message or undef
sub CLI_VALIDATE ( $self, $opt, $arg, $rest ) {
    return;
}

sub CLI_RUN ( $self, $opt, $arg, $rest ) {
    return;
}

1;
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::Core::CLI::Cmd

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 ATTRIBUTES

=head1 METHODS

=head1 SEE ALSO

=cut
