package Pcore::App::API::Backend::remote;

use Pcore -class, -res;

with qw[
  Pcore::App::API::Backend
];

sub init () {
    return res [ 500, 'Not implemented' ];
}

1;
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::App::API::Backend::remote

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 ATTRIBUTES

=head1 METHODS

=head1 SEE ALSO

=cut
