package Pcore::Util::List;

use Pcore -export;
use List::AllUtils qw[:all];    ## no critic qw[Modules::ProhibitEvilModules]

our $EXPORT = [qw[pairs pairkeys pairvalues]];

1;
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::Util::List

=head1 SYNOPSIS

=head1 DESCRIPTION

=cut
